#!/usr/bin/env python3
import numpy as np
import pandas as pd
import os

# import re
import argparse
import csv
import json
import yaml
import gzip

import click


@click.command(
    help="Make a pretty table",
    epilog="You need to open the CSV in a spreadsheet that understands linebreaks",
)
@click.option(
    "-m",
    "--vocdir",
    metavar="DIR",
    required=False,
    default=None,
    type=str,
    help="directory containing the yamls defining the variant of concerns",
)
@click.option(
    "-a",
    "--amplicons",
    "amp",
    metavar="YAML",
    required=False,
    default=None,
    type=str,
    help="list of query amplicons, from mutbamscan",
)
# TODO: add mutually exclusive options
@click.option(
    "-j",
    "--json",
    "json_fname",
    metavar="JSON",
    type=str,
    help="results generated by mutbamscan",
)
@click.option(
    "-y",
    "--yaml",
    "yaml_fname",
    metavar="YAML",
    type=str,
    help="results generated by mutbamscan",
)
@click.option(
    "-o",
    "--output",
    "csv_fname",
    metavar="CSV",
    required=False,
    default="scanned_article.csv",
    type=str,
    help="name of (pretty) csv file to save the table into",
)
# TODO: add mutually exclusive options
@click.option(
    "-e/-E",
    "--escape/--no-escape",
    "escape",
    default=False,
    help="use escape characters for newlines",
)
@click.option(
    "-x",
    "--excel",
    "semi",
    is_flag=True,
    default=False,
    help="use a semi-colon ';' instead of a comma ',' in the comma-separated-files as required by Microsoft Excel",
)
@click.option(
    "--batchname",
    "batchname",
    metavar="SEP",
    required=False,
    default=None,
    type=str,
    help="separator used to split samplename/batchname in separate column",
)
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    default=False,
    help="Run quietly: do not print the table",
)
def cooc_pubmut(
    vocdir, amp, json_fname, yaml_fname, csv_fname, escape, semi, batchname, quiet
):
    escape = (
        ("\\n", csv.QUOTE_NONNUMERIC, "\\")
        if escape
        else ("\n", csv.QUOTE_MINIMAL, None)
    )

    # TODO make the "header prettification" code more generic and share with colourmut

    # load variants
    variants_names = {}
    if vocdir is not None:
        for path in [p for p in os.listdir(vocdir) if not p.startswith(".")]:
            full_path = os.path.join(vocdir, path)
            with open(full_path, "r") as yf:
                loaded_yaml = yaml.load(yf, Loader=yaml.FullLoader)["variant"]
            nameparts = []
            # lineage name
            if "pangolin" in loaded_yaml:
                nameparts += [loaded_yaml["pangolin"]]
            elif "nextstrain" in loaded_yaml:
                nameparts += [loaded_yaml["nextstrain"]]
            # WHO greek letter
            if "who" in loaded_yaml:
                nameparts += [f"({loaded_yaml['who']})"]
            # assemble, e.g. "B.1.529 (omicron)"
            variants_names[loaded_yaml["short"]] = " ".join(nameparts)
        # print(variants_names)

    # load amplicons
    amplicon_nfo = {}
    if amp is not None:
        assert os.path.isfile(amp), f"cannot find amplicon file yaml file {amp}"
        with open(amp, "rt") as yf:
            amp_str = yaml.safe_load(yf)

        amplicon_nfo = {
            a: "\n".join(
                [
                    # amplicon number
                    f"Amplicon {a.split('_')[0]}",
                    # genomic position span
                    f"[{aqu[0]}-{aqu[1]}]",
                    "",
                    # Variants
                    ",".join([variants_names.get(v, v) for v in a.split("_")[1:]]),
                    "",
                    # Mutations
                    # TODO incorporate output of curate
                    ",".join(
                        [
                            f"{p}{b}"
                            if len(b) == 1
                            else (
                                f"\u0394{p}-{p + len(b) - 1}"
                                if b == "-" * len(b)
                                else f"{p}\u2192{b}"
                            )
                            for p, b in aqu[4].items()
                        ]
                    ),
                ]
            )
            for a, aqu in amp_str.items()
        }
        # print(amplicon_nfo)

    # load table
    table = {}

    assert not (
        semi and csv_fname.rfind(".tsv") != -1
    ), f"Excel cannot use TSV files {csv_fname}"

    if json_fname:
        assert os.path.isfile(json_fname), f"cannot find result json file {json_fname}"
        with open(json_fname, "rt") as jf:
            table = json.load(fp=jf)
    elif yaml_fname:
        assert os.path.isfile(yaml_fname), f"cannot find result json file {yaml_fname}"
        with open(yaml_fname, "rt") as yf:
            table = yaml.safe_load(yf)

    assert len(table) > 0, "cannot succesfully load table"

    #
    # pretty output for article
    #

    df_dict = {}
    for sam, amplicons in table.items():
        print(sam)
        # table key
        ksam = None
        if batchname:
            (sam, ignore, batch) = sam.rpartition(batchname)
            ksam = (sam, batch) if sam and batch else sam or batch
        else:
            ksam = sam
        df_dict[ksam] = {}

        for ampname, amp in amplicons.items():
            # get topmost
            sites_cnt_l = -1
            sites_cnt = 0
            if amp["sites"]:  # empty ?
                (sites_cnt_l, sites_cnt) = list(amp["sites"].items())[-1]
            muts_cnt_l = -1
            muts_cnt = 0
            if amp["muts"]:  # empty ?
                (muts_cnt_l, muts_cnt) = list(amp["muts"].items())[-1]

            if int(muts_cnt_l) < int(sites_cnt_l):
                muts_cnt = 0

            # pack into dict for pandas
            df_dict[ksam].update(
                {
                    amplicon_nfo.get(
                        ampname, ampname
                    ): f"{muts_cnt} / {sites_cnt}{escape[0]}{f'{100*float(muts_cnt)/float(sites_cnt) :.2f}%' if sites_cnt else 'NA'}"
                }
            )

    pretty_table_df = pd.DataFrame.from_dict(data=df_dict, orient="index")
    # TODO rename column with pretty names. (like in colourmut)
    if not quiet:
        with pd.option_context(
            "display.max_rows", None
        ):  # , 'display.max_columns', None):
            print(pretty_table_df)
    pretty_table_df.to_csv(
        csv_fname,
        quoting=escape[1],
        escapechar=escape[2],
        sep=("\t" if csv_fname.rfind(".tsv") != -1 else ";" if semi else ","),
        compression={"method": "infer"},
    )


if __name__ == "__main__":
    cooc_pubmut()
